module Language.SystemU.Coerce where

import Language.SystemU.Syntax hiding (Var)
import qualified Language.SystemU.Syntax as S
import qualified Language.SystemU.Builtins as B
import Language.SystemU.Sorted (Sorted, Sort(..), sort)
import qualified Language.SystemU.Sorted as Sorted
import qualified Language.SystemU.TypeCheck as T


import Prelude hiding (pi)
import Control.Applicative
import Control.Monad (when)
import System.IO.Unsafe

data U = Var (Tm ())
       | ProdStar U U
       | ProdBox String (Tm ()) U
       | ProdDelta String U
  deriving (Show)       
         
instance Eq U where
  a == b = unsafePerformIO $ aeqU a b

-- TODO should we normalize here? It's not alpha equivalence. it's beta!
-- just a hack for now...
aeqU :: U -> U -> IO Bool
aeqU (Var x) (Var y) = aeq (norm x) (norm y)  
aeqU (ProdStar l1 r1) (ProdStar l2 r2) = 
  (&&) <$> aeqU l1 l2 <*> aeqU r1 r2
aeqU (ProdBox s1 t1 b1) (ProdBox s2 t2 b2) = do
  s <- genSym
  (&&) <$> aeq t1 t2 <*> aeqU (renameU s1 s b1) (renameU s2 s b2)
aeqU (ProdDelta s1 b1) (ProdDelta s2 b2) = do 
  s <- genSym
  aeqU (renameU s1 s b1) (renameU s2 s b2)

-- Must be in beta normal form and of kind *.
-- TODO: merge with QuoteType.quoteType
tyU :: Sorted -> U
tyU ty = 
  let go = tyU
  in  
  case ty of
    (Pi Type x l r) 
      | sort l == Type    -> ProdStar (go l) (go r)
      | sort l == Kind    -> ProdBox x (erase l) (go r)
      | sort l == BoxSort -> ProdDelta x (go r)
    _ -> Var (erase ty)

unU :: U -> Tm ()
unU (Var tm)         = tm
unU (ProdStar a b)   = unU a ==> unU b
unU (ProdBox nm t u) = pi nm t $ unU u
unU (ProdDelta nm u) = pi nm box $ unU u

compileU :: U -> Tm ()
compileU = norm . go
  where 
    go u =
      case u of 
        Var tm          -> app B._Var tm
        ProdStar a b    -> app2 B._ProdStar (go a) (go b)
        ProdBox nm k u  -> app2 B._ProdBox k $ lam nm k (go u)
        ProdDelta nm u  -> app B._ProdDelta $ lam nm box (go u)

data C = CoerceProdStar U U U U C C
       | CoerceProdBox String (Tm ()) U U C
       | CoerceProdDelta String U U C         
       | ReifyProdStar U U C C
       | ReflectProdStar U U C C
       | ReifyProdBox String (Tm ()) U C
       | ReflectProdBox String (Tm ()) U C
       | ReifyProdDelta String U C
       | ReflectProdDelta String U C
       | Identity (Tm ())
  deriving (Eq, Show)       
       


reify :: U -> C         
reify (Var tm) = Identity tm
reify (ProdStar u1 u2) = ReifyProdStar u1 u2 (reflect u1) (reify u2)
reify (ProdBox nm k u) = ReifyProdBox nm k u (reify u)
reify (ProdDelta nm u) = ReifyProdDelta nm u (reify u)
         
reflect :: U -> C
reflect (Var tm) = Identity tm
reflect (ProdStar u1 u2) = ReflectProdStar u1 u2 (reify u1) (reflect u2)
reflect (ProdBox nm k u) = ReflectProdBox nm k u (reflect u)
reflect (ProdDelta nm u) = ReflectProdDelta nm u (reflect u)

coerce :: U -> U -> C
coerce (Var a')  (Var b') 
  | a == b    = Identity a
  | otherwise = error $ unlines ["coerce Var Var: ", show a, show b]
  where a = norm a'
        b = norm b'
coerce (Var _) u = reify u
coerce u (Var _) = reflect u
coerce (ProdStar a1 b1) (ProdStar a2 b2) =
  CoerceProdStar a1 b1 a2 b2 (coerce a2 a1) (coerce b1 b2)
  
coerce (ProdBox nm1 k1 u1) (ProdBox nm2 k2 u2) = unsafePerformIO $ do
  n <- genSym
  let u1' = renameU nm1 n u1
      u2' = renameU nm2 n u2
  return $ CoerceProdBox n k1 u1' u2' (coerce u1' u2')

coerce (ProdDelta nm1 u1) (ProdDelta nm2 u2) = unsafePerformIO $ do
  n <- genSym
  let u1' = renameU nm1 n u1
      u2' = renameU nm2 n u2
  return $ CoerceProdDelta n u1' u2' (coerce u1' u2')
    
coerce u1 u2 = error $ unlines [ "coerce:"
                               , "u1 = " ++ show u1
                               , "u2 = " ++ show u2
                               ]  

renameU :: String -> String -> U -> U
renameU s1 s2 u =
  let goU = renameU s1 s2
      go  = rename s1 s2
  in
  case u of 
    Var x             -> Var $ go x
    ProdStar u1 u2    -> ProdStar (goU u1) (goU u2)
    ProdBox s t u 
      | s == s1       -> ProdBox s (go t) u
      | s == s2       -> error "renameU ProdBox: not fresh!"                   
      | otherwise     -> ProdBox s (go t) (goU u)                 
    ProdDelta s u 
      | s == s1       -> ProdDelta s u
      | s == s2       -> error "renameU ProdDelta: not fresh!"                   
      | otherwise     -> ProdDelta s (goU u)                 
    
substU :: String -> Tm () -> U -> IO U
substU nm ty u =
  let goU = substU nm ty 
      go  = subst nm ty 
  in
  case u of
    Var x             -> Var <$> go x
    ProdStar u1 u2    -> ProdStar <$> goU u1 <*> goU u2
    ProdBox nm' t u'  
      | nm == nm'     -> ProdBox nm' <$> go t <*> pure u'
      | otherwise     -> do -- rename nm' to avoid capture
                            s <- genSym
                            let u'' = renameU nm' s u'
                            ProdBox s <$> go t <*> goU u''
    ProdDelta nm' u'  -> do s <- genSym                        
                            let u'' = renameU nm' s u'
                            ProdDelta s <$> goU u''
                            
assert :: String -> String -> Bool -> Either String ()
assert ctx msg b = 
  when (not b) $ Left $ 
    "[" ++ ctx ++ "] assertion failed: " ++ msg
    
-- returns a pair of (input,output) types of the coercion
typecheckC :: C -> Either String (U,U)
typecheckC (Identity tm) = Right (Var tm, Var tm)
typecheckC (CoerceProdStar l1 r1 l2 r2 cl cr) = do
  let assert' = assert "typecheckC/CoerceProdStar"
  (l2',l1') <- typecheckC cl
  (r1',r2') <- typecheckC cr
  assert' (unlines ["l1 == l1'", show l1, show l1', show cl]) (l1 == l1')
  assert' "l1 == l1'" (l1 == l1')
  assert' (unlines ["l2 == l2'" , show l2, show l2', show cl]) (l2 == l2')
  assert' "l2 == l2'" (l2 == l2')
  assert' "r1 == r1'" (r1 == r1')
  assert' "r2 == r2'" (r2 == r2')
  return (ProdStar l1 r1, ProdStar l2 r2)
typecheckC (CoerceProdBox nm k u1 u2 c) = do
  let assert' = assert "typecheckC/CoerceProdBox"
  (u1',u2') <- typecheckC c
  assert' "u1 == u1'" (u1 == u1')
  assert' "u2 == u2'" (u2 == u2')
  return (ProdBox nm k u1, ProdBox nm k u2)
typecheckC (CoerceProdDelta nm u1 u2 c) = do
  let assert' = assert "typecheckC/CoerceProdDelta"
  (u1',u2') <- typecheckC c
  assert' "u1 == u1'" (u1 == u1')
  assert' "u2 == u2'" (u2 == u2')
  return (ProdDelta nm u1, ProdDelta nm u2)

typecheckC (ReifyProdStar u1 u2 c1 c2) = do
  let assert' = assert "typecheckC/ReifyProdStar"
  let t1 = unU u1    
      t2 = unU u2
  (u1',Var t1') <- typecheckC c1
  (Var t2',u2') <- typecheckC c2
  assert' "u1 == u1'" (u1 == u1')
  assert' "u2 == u2'" (u2 == u2')
  assert' "t1 == t1'" (t1 == t1')
  assert' "t2 == t2'" (t2 == t2')
  return (Var $ t1 ==> t2, ProdStar u1 u2)
typecheckC (ReflectProdStar u1 u2 c1 c2) = do
  let assert' = assert "typecheckC/ReflectProdStar"
  let t1 = unU u1    
      t2 = unU u2
  (Var t1',u1') <- typecheckC c1
  (u2',Var t2') <- typecheckC c2
  assert' "u1 == u1'" (u1 == u1')
  assert' "u2 == u2'" (u2 == u2')
  assert' "t1 == t1'" (t1 == t1')
  assert' "t2 == t2'" (t2 == t2')
  return (ProdStar u1 u2, Var $ t1 ==> t2)  
typecheckC (ReifyProdBox nm k u c) = do
  let assert' = assert "typecheckC/ReifyProdBox"
  let t = unU u
  (Var t', u') <- typecheckC c    
  assert' "t == t'" (t == t')
  assert' "u == u'" (u == u')
  return (Var $ pi nm k t, ProdBox nm k u)
typecheckC (ReflectProdBox nm k u c) = do
  let assert' = assert "typecheckC/ReflectProdBox"
  let t = unU u
  (u',Var t') <- typecheckC c    
  assert' "t == t'" (t == t')
  assert' "u == u'" (u == u')
  return (ProdBox nm k u, Var $ pi nm k t) 
typecheckC (ReifyProdDelta nm u c) = do
  let assert' = assert "typecheckC/ReifyProdDelta"
  let t = unU u
  (Var t', u') <- typecheckC c    
  assert' "t == t'" (t == t')
  assert' "u == u'" (u == u')
  return (Var $ pi nm box t, ProdDelta nm u)
typecheckC (ReflectProdDelta nm u c) = do
  let assert' = assert "typecheckC/ReflectProdDelta"
  let t = unU u
  (u',Var t') <- typecheckC c    
  assert' "t == t'" (t == t')
  assert' "u == u'" (u == u')
  return (ProdDelta nm u, Var $ pi nm box t) 

-- generate a coercion term. coercion functions are
-- free variables.
  
compileCoercion :: Tm () -> C -> Tm ()
compileCoercion v = go
  where
    [cS,cB,cD] = map (\s -> var $ "coerceExp_" ++ [s]) "*#^"
    [rS,rB,rD] = map (\s -> var $ "reifyProd_" ++ [s]) "*#^"
    [fS,fB,fD] = map (\s -> var $ "reflectProd_" ++ [s]) "*#^"      

    cmpU = compileU
    cmpUB nm k = lam nm k . cmpU
    cmpUD nm   = lam nm box . cmpU
    goB nm k  = lam nm k . go
    goD nm    = lam nm box . go
    go c = 
      case c of
        CoerceProdStar l1 r1 l2 r2 c1 c2 ->
          let [a1,b1,a2,b2] = map cmpU [l1,r1,l2,r2] in
            app7 cS v a1 b1 a2 b2 (go c1) (go c2)
        CoerceProdBox nm k u1 u2 c ->
          let [t1,t2] = map (cmpUB nm k) [u1,u2] in
          app5 cB v k t1 t2 (goB nm k c)
        CoerceProdDelta nm u1 u2 c ->
          let [t1,t2] = map (cmpUD nm) [u1,u2] in
          app4 cD v t1 t2 (goD nm c) 
        ReifyProdStar u1 u2 c1 c2 ->
          let [t1,t2] = map cmpU [u1,u2] in
          app5 rS t1 t2 v (go c1) (go c2)
        ReflectProdStar u1 u2 c1 c2 ->
          let [t1,t2] = map cmpU [u1,u2] in
          app5 fS t1 t2 v (go c1) (go c2)
        ReifyProdBox nm k u c ->
          app4 rB k (cmpUB nm k u) v (goB nm k c)
        ReflectProdBox nm k u c ->
          app4 fB k (cmpUB nm k u) v (goB nm k c)
        ReifyProdDelta nm u c ->
          app3 rD (cmpUD nm u) v (goD nm c)
        ReflectProdDelta nm u c ->
          app3 fD (cmpUD nm u) v (goD nm c)
        Identity t -> app2 (var "reifyVar") v t

-- for a (#,*) product type (Pi nm:k. t) and a type 
-- parameter a:k, build a coercion with the type
-- PExp V ([a/nm][t]) -> PExp V [[a/nm]t]
buildCoercion :: T.Env -> Tm () -> Tm () -> Tm () -> IO (Tm ())
buildCoercion env v prod' a' = do
  let prod = norm prod'
      a    = norm a'
  let Just (Pi _ nm _ t) = T.typedSortInf env prod
  u1 <- substU nm a (tyU t)
  t1 <- fmap norm $ subst nm a (erase t)
  let Just t2 = T.typedSortInf env t1
      u2 = tyU t2
      c = coerce u1 u2
  (u1',u2') <- either fail return $ typecheckC c

  when (u1 /= u1') $ fail "buildCoercion: u1 /= u1'"
  when (u2 /= u2') $ fail "buildCoercion: u2 /= u2'"  
  
  validateCoercion env v c 
  let cTm  = compileCoercion v c  
  return cTm
  
localizeCoercionFailure env v c =
  let go = validateCoercion env v
      goWith nm t = validateCoercion ((nm,t):env) v
  in    
  case c of 
    CoerceProdStar _ _ _ _ c1 c2 -> go c1 >> go c2
    CoerceProdBox nm k _ _ c     -> goWith nm k c
    CoerceProdDelta nm _ _ c     -> goWith nm box c
    ReifyProdStar _ _ c1 c2      -> go c1 >> go c2
    ReflectProdStar _ _ c1 c2    -> go c1 >> go c2
    ReifyProdBox nm k _ c        -> goWith nm k c
    ReflectProdBox nm k _ c      -> goWith nm k c
    ReifyProdDelta nm _ c        -> goWith nm box c
    ReflectProdDelta nm _ c      -> goWith nm box c
    Identity _                   -> return ()

validateCoercion env v c = do  
  (u1,u2) <- either fail return $ typecheckC c
  let cTm  = compileCoercion v c  
      cU1 = compileU u1
      cU2 = compileU u2
      vars = tmVal $ fvs cTm 
      S.Var _ vNm = v
      ty = norm $ app2 B._PExp v cU1 ==> app2 B._PExp v cU2
  cTy <- 
    case T.typecheck ([(vNm,B.u ==> star)] ++ env ++ T.coerceEnv) cTm of 
      Left e     -> fail $ unlines [
                       "buildCoercion: coercion did not type check"
                     , e
                     , show vars                       
                     , show c
                     ]  
      Right cTm' -> return $ tmVal cTm'  
      
  when (cTy /= ty) $ do
    localizeCoercionFailure env v c 

    -- if we get back here, children did not fail
    fail $ unlines [
        "buildCoercion: compiled coercion type error"
{--}
      , "got " ++ show cTy
      , "expected " ++ show ty
--}
      , "u1 = " ++ show u1
      , "u2 = " ++ show u2  
      , "c  = " ++ show c  
      ]  


