module Language.SystemU.Core where

import qualified Data.Map as M
import Data.Map (Map)
import qualified Data.Set as S
import Data.Set (Set)
import Control.Monad.Reader
import Control.Monad.State
import Control.Monad.Error
import Control.Monad
import Control.Applicative
import Prelude hiding (exp)
import System.IO.Unsafe

-- Phases:
-- Desugar (remove decls) : Pgm -> Tm
-- Sort-checking and elaboration
-- (optional) type check inside type quotations
-- expand type quotations
-- (optional) type check inside term quotations
-- type check and expand term quotations
-- evaluation

data Stmt = Stmts [Stmt]
          | SDecl String Tm Tm
  deriving (Read)

data Pgm = Pgm Stmt Tm
  deriving (Read)

data Tm = Star | Box | Delta 
        | Var String
        | Lam String Tm Tm
        | Pi String Tm Tm
        | App Tm Tm
        | Decl String Tm Tm Tm
        | Quote Tm  
  deriving (Ord, Read, Show)
           
instance Eq Tm where           
  e1 == e2 = e1 `alphaEq` e2
  
traceError msg e =  
  e `catchError` (\s -> throwError $ msg ++ "\n" ++ s)
  
alphaEq :: Tm -> Tm -> Bool  
alphaEq e1 e2 = evalState (aeq e1 e2) 0

fresh :: State Int String
fresh = do
  i <- get
  put (i+1)
  return (show i)

aeq :: Tm -> Tm -> State Int Bool
aeq Star Star   = return True
aeq Box Box     = return True
aeq Delta Delta = return True
aeq (App l1 r1) (App l2 r2) = (&&) <$> aeq l1 l2 <*> aeq r1 r2
aeq (Pi n1 t1 b1) (Pi n2 t2 b2) = do
  n <- fresh
  let b1' = substitute n1 (Var n) b1
      b2' = substitute n2 (Var n) b2
  (&&) <$> aeq t1 t2 <*> aeq b1' b2'
aeq (Lam n1 t1 b1) (Lam n2 t2 b2) = do
  n <- fresh
  let b1' = substitute n1 (Var n) b1
      b2' = substitute n2 (Var n) b2
  (&&) <$> aeq t1 t2 <*> aeq b1' b2'
aeq (Var x) (Var y) = return $ x == y
aeq (Decl n1 t1 d1 b1) (Decl n2 t2 d2 b2) = do
  n <- fresh
  let sub nm = substitute nm (Var n)
  and <$> sequence [aeq t1 t2, aeq d1 d2, aeq (sub n1 b1) (sub n2 b2)] 

aeq t1 t2 = return False
            --fail $ unlines ["aeq", paren $ show t1, paren $ show t2]

paren s = '(' : s ++ ")"


rules = M.fromList [
    ((Star,Star), Star), ((Box, Star), Star), ((Delta, Star), Star)
  , ((Box, Box), Box), ((Delta, Box), Box)                           
  ]              
        
type Env = Map String Tm     
type Tc = ReaderT Env (Either String)

typecheckPgm :: Pgm -> Either String Tm
typecheckPgm = typecheck . pgm2Tm

pgm2Tm (Pgm stmt tm) = compileStmt stmt tm
  
compileStmt :: Stmt -> Tm -> Tm
compileStmt (Stmts stmts) = foldl (.) id $ map compileStmt stmts
compileStmt (SDecl nm ty bnd) = Decl nm ty bnd

typecheck :: Tm -> Either String Tm
typecheck t = runReaderT (traceError "typecheck" $ tc t) M.empty

tc :: Tm -> Tc Tm
tc Star = return Box
tc Box  = return Delta
tc (Pi nm ty body) = traceError ("tc Pi " ++ nm) $ do
  s1 <- tc ty
  s2 <- withReaderT (M.insert nm ty) $ tc body
  let msg = "Error: tc Pi"
  maybe (throwError msg) return $ M.lookup (s1,s2) rules
tc (App e1 e2) = do
  t1 <- traceError "tcApp e1" $ tc e1
  t2 <- traceError ("tcApp e2 = " ++ show e2) $ tc e2
  t1' <- traceError "tc App: normalize t1" $ normalize t1
  t2' <- traceError "tc App: normalize t2" $ normalize t2
  case t1' of
    Pi nm l r   | l == t2' -> normalize $ substitute nm e2 r
                | otherwise -> do e1' <- normalize e1
                                  e2' <- normalize e2
                                  throwError $ unlines [
                                      "App: type mismatch"
                                    , "e1: " ++ show e1', ""
                                    , "t1: " ++ show t1, ""
                                    , "e2: " ++ show e2', ""
                                    , "l: " ++ show l, ""
                                    , "t2': " ++ show t2', ""
                                    ]
    _                       -> throwError $ unlines ["App: not a Pi type", show t1, show e1, show e2]
  
tc (Lam nm ty body) = traceError ("tc Lam " ++ nm) $ do  
  s <- traceError ("tc Lam " ++ nm ++ " ty") $ tc ty
  ty2 <- traceError ("tc Lam " ++ nm ++ " body") $
         withReaderT (M.insert nm ty) $ tc body
  return (Pi nm ty ty2)
tc (Var nm) = do
  let msg = "Error: tc Var " ++ nm
  maybe (throwError msg) norm =<< asks (M.lookup nm)
tc (Decl nm ty1 bnd body) = traceError ("tc Decl " ++ nm) $ do  
  let msg s = "tc Decl " ++ nm ++ " / " ++ s
  ty1' <- traceError (msg "normalize ty1") $ normalize ty1
  ty2 <- traceError (msg "tc bnd") $ tc bnd
  ty2' <- traceError (msg "normalize ty2") $ normalize ty2
  when (ty1' /= ty2') $ throwError "Decl: type mismatch"
  
  tyTy2 <- traceError (msg "tc ty2") $ tc ty2
  -- if bnd is not a term, normalize it.
  {--
  bnd' <- if tyTy2 == Star 
            then return bnd 
            else do --bnd' <- expandTy bnd
                    traceError (msg "normalize bnd'") $
                      normalize bnd
  --}
  let body' = substitute nm bnd body
  --traceError (msg "tc substituted body: "  ++ nm ++ "=" ++ show bnd') $ tc body'
  traceError (msg "tc substituted body") $ tc body'

tc (Quote e) = traceError "tc Quote" $ do
  t <- tc e
  case t of
    Star -> return u
{--}
    _    -> do Star <- tc t
               qt <- quoteTy =<< normalize t
               u' <- tc qt
               when (u' /= u) $ throwError $ "Quote type error: " ++ unlines [show u',show u]
               let t' = App pExp qt
               Star <- tc t'
               return t'
--}
    
-- built-ins    
u, exp, pExp, var, prodStar, prodBox, prodDelta :: Tm
[u,exp,pExp,var,prodStar,prodBox,prodDelta] = map load nms
  where load nm = pgm2Tm $ read $ unsafePerformIO $ readFile $ "examples/pgms/" ++ nm
        nms = words "U Exp PExp Var ProdStar ProdBox ProdDelta"

mkRep,mkAbsStar,mkAppStar,mkAbsBox,mkAppBox,mkAbsDelta,mkAppDelta :: Tm
[mkRep,mkAbsStar,mkAppStar,mkAbsBox,mkAppBox,mkAbsDelta,mkAppDelta] = map load nms
  where load nm = pgm2Tm $ read $ unsafePerformIO $ readFile $ "examples/pgms/" ++ nm
        nms = words "mkRep mkAbsStar mkAppStar mkAbsBox mkAppBox mkAbsDelta mkAppDelta"

_DistExpStar :: Tm
[_DistExpStar, _DistExpBox, _DistExpDelta, 
 distExpStar, distExpBox, distExpDelta, 
 _FactorExpStar, _FactorExpBox, _FactorExpDelta,
 factorExpStar, factorExpBox, factorExpDelta,
 _CoerceExpStar, _CoerceExpBox, _CoerceExpDelta,
 coerceExpStar, coerceExpBox, coerceExpDelta,
 
 _ReifyVar, _ReifyProdStar, _ReifyProdBox, _ReifyProdDelta,
 reifyVar, reifyProdStar, reifyProdBox, reifyProdDelta,
 _ReflectProdStar, _ReflectProdBox, _ReflectProdDelta,
 reflectProdStar, reflectProdBox, reflectProdDelta
 ] = map load nms
  where load nm = pgm2Tm $ read $ unsafePerformIO $ readFile $ "examples/pgms/" ++ nm
        nms = words $ unwords [
                "DistExpStar DistExpBox DistExpDelta" 
               ,"distExpStar distExpBox distExpdelta" 
               ,"FactorExpStar FactorExpBox FactorExpDelta" 
               ,"factorExpStar factorExpBox factorExpDelta" 
               ,"CoerceExpStar CoerceExpBox CoerceExpDelta"
               ,"coerceExpStar coerceExpBox coerceExpDelta"
               ,"ReifyVar ReifyProdStar ReifyProdBox ReifyProdDelta"
               ,"reifyVar reifyProdStar reifyProdBox reifyProdDelta"
               ,"ReflectProdStar ReflectProdBox ReflectProdDelta"
               ,"reflectProdStar reflectProdBox reflectProdDelta"
               ]


-- quote a normalized type of kind *
quoteTy :: Tm -> Tc Tm        
quoteTy ty = do
  Star <- tc ty -- sanity check!
  case ty of
    (Pi nm l r) -> do tl <- tc l 
                      traceError ("quoteTy / Pi " ++ nm) $
                       withReaderT (M.insert nm l) $ do                      
                        Star <- tc r   -- sanity check!
                        q <- case tl of
                          Star  -> app2 prodStar <$> quoteTy l <*> quoteTy r
                          Box   -> app2 prodBox      l         <$> (Lam nm l <$> quoteTy r)
                          Delta -> App prodDelta               <$> (Lam nm l <$> quoteTy r)
                        u' <- tc q
                        when (u' /= u) $ throwError "quoteTy: result not of kind u!"
                        
                        -- norm here is necessary!!! WTF???
                        q' <- norm q
                        u' <- tc q'  
                        when (u' /= u) $ throwError "quoteTy: result not of kind u!"
                        
                        return q'
    _           -> return $ App var ty

app2 a b c = App (App a b) c

-- expand type quotations 
expandTy :: Tm -> Tc Tm
expandTy e = 
  case e of
    Quote e    -> do t <- tc e 
                     throwError $ "expandTy " ++ show e
                     if t == Star then quoteTy e else throwError "expandTy: not a type"
    _          -> return e

normalize' :: Tm -> Tm
normalize' t = fromRight $ runReaderT (normalize t) M.empty
  where fromRight (Right a) = a

normalize :: Tm -> Tc Tm
normalize  = norm

norm :: Tm -> Tc Tm
norm e = do
  v <- eval e
  case v of
    Star       -> return Star
    Box        -> return Box
    Delta      -> return Delta
    Lam nm t b -> Lam nm <$> norm t <*> norm b
    Pi nm t b  -> Pi nm  <$> norm t <*> norm b
    Quote e'   -> do t <- norm =<< tc e'
                     case t of
                       -- e' is a type
                       Star -> do qe <- quoteTy e'
                                  traceError ("norm / Quote: e' = " ++ show e' ++ ", qe = " ++ show qe ++ " checking ") $ do
                                    u' <- tc qe
                                    when (u' /= u) $ throwError $ "Quote type error: " ++ unlines [show u',show u]
                                  norm qe
                       _    -> return v -- throwError $ "norm / Quote: bad type: " ++ show t
                     
    Decl _ _ _ _ -> throwError ("norm Decl")                 
    App a b    -> App <$> norm a <*> norm b
    Var _      -> return v

evaluate :: Tm -> Tc Tm
evaluate = eval

eval (App t1 t2) = do
  t1' <- eval t1  
  case t1' of
    Lam nm t body -> eval $ substitute nm t2 body
    _             -> return $ App t1' t2
    
eval e@(Var _) = return e
eval e@(Lam nm t b) = Lam nm <$> eval t <*> return b
eval (Pi nm t b) = Pi nm <$> eval t <*> eval b
eval Star = return Star
eval Box  = return Box
eval Delta = return Delta

eval (Quote e) = do 
  t <- normalize =<< tc e
  case t of
    -- e' is a type
    Star -> do qe <- quoteTy e
               traceError ("eval / Quote: e' = " ++ show e ++ ", qe = " ++ show qe ++ " checking ") $ do
                 u' <- tc qe
                 when (u' /= u) $ 
                   throwError $ "Quote type error: " ++ unlines [show u',show u]
                 eval qe
    _    -> return $ Quote e -- throwError $ "norm / Quote: bad type: " ++ show t

eval (Decl nm ty1 bnd body) = eval $ substitute nm bnd body

substitute :: String -> Tm -> Tm -> Tm
substitute = subst

subst nm v = go
  where go e@(Var nm') | nm == nm' = v
                       | otherwise = e
        go (App t1 t2) = App (go t1) (go t2)
        go e@(Lam nm' t b) | nm == nm' = Lam nm' (go t) b
                           | otherwise = Lam nm' (go t) (go b)
        go (Pi nm' t b) | nm == nm' = Pi nm' (go t) b  
                        | otherwise = Pi nm' (go t) (go b)
        go (Decl nm' t d b) | nm == nm' = Decl nm' (go t) (go d) b
                            | otherwise = Decl nm' (go t) (go d) (go b)
        go (Quote t) = Quote (go t)
        go Star = Star
        go Box = Box
        go Delta = Delta