module Language.SystemU.TypeCheck where

import Language.SystemU.Syntax hiding (app)
import qualified Language.SystemU.Syntax as S
import Language.SystemU.Sorted hiding (Env)
import qualified Language.SystemU.Sorted as Sorted
import Language.SystemU.QuoteType
import Language.SystemU.Builtins as B

import Control.Applicative
import Control.Monad (when)
import Control.Arrow (second)

import Prelude hiding (pi)
import System.IO.Unsafe

type Typed = Tm (Tm ())
type Kinded = Tm (Tm ())

type Env = [(String,Tm ())]

tmType = tmVal

typecheckTypes :: Sorted -> Either String ()
typecheckTypes = typecheckTypes' []
  
typecheckTypes' :: Env -> Sorted -> Either String ()
typecheckTypes' env e =
  let go = typecheckTypes' env  
      goWith x t e = typecheckTypes' ((x, erase t):env) e
  in    
  case e of
    Var _ _        -> return ()
    App _ f a      -> go f >> go a
    Lam _ x t b    -> go t >> goWith x t b
    Pi _ x t b     -> go t >> goWith x t b
    Decl _ x t d b -> go t >> go d >> go (unsafePerformIO $ subst x d b)

    Quote Type t   -> do -- only quote types of kind *
                         k <- typecheck env t
                         if isStar k
                           then return ()
                           else Left "Quoted types must have kind *"

    Quote Term t   -> go t
    Star _         -> return ()
    Box _          -> return ()
    Delta _        -> return ()
    _              -> Left $ "typecheckTypes: " ++ show e    


rules = 
  let star  = Star ()
      box   = Box ()
      delta = Delta ()
  in [
    ((star,star), star), ((box, star), star), ((delta, star), star)
  , ((box, box), box), ((delta, box), box)                           
  ]              

app :: Typed -> Typed -> Either String Typed
app f a = do
  let tf = norm $ tmType f
      ta = norm $ tmType a
      ta' = erase ta
  case tf of
    Pi s x l r | l' == ta' -> let t = norm $ unsafePerformIO $ subst x (erase a) r
                              in return $ App t f a
               | otherwise -> let Left (dl,da) = unsafePerformIO $ runDiff l' ta' in
{--      
                              Left $ unlines [
                                 "appTyped: type mismatch"
                               , "l  = " ++ show l  
                               , "ta = " ++ show ta
                               , "dl = " ++ show dl
                               , "da = " ++ show da  
                               , "f  = " ++ show (erase f)
                               , "a  = " ++ show (erase a)
                               ]  
--}
                              Left $ unlines [
                                 "app: argument type mismatch"
                               , "expected  = " ++ show l  
                               , "but got   = " ++ show ta
                               , "in the application of: "   
                               , "    " ++ show (erase f)
                               , "to: " ++ show (erase a)
                               ]  
               where l' = erase l                         
    _                                -> Left $ unlines [
                                           "appTyped: not a pi type: " ++ show tf
                                         , "f = " ++ show (erase f)
                                         , "a = " ++ show (erase a)
                                         ]  

-- invariant: always return a type in normal form
typecheck :: Show a => Env -> Tm a -> Either String Typed
typecheck env e =
  let go = typecheck env
      goWith x e = typecheck ((x, norm e):env)
  in    
  case e of 
    Var _ x      -> do let failure = Left $ "unknown variable: " ++ x
                       t <- maybe failure return $ lookup x env
                       return $ Var t x
    Pi _ x t b   -> do t' <- go t
                       b' <- goWith x (erase t) b
                       let failure = Left $ unlines [
                                        "Error: typecheck Pi: " ++ show e
                                      , show (s1,s2)  
                                      ]  
                           s1      = tmType t'
                           s2      = tmType b'
                       s <- maybe failure return $ lookup (s1,s2) rules
                       return $ Pi s x t' b'
    App _ f a    -> do f' <- go f
                       a' <- go a
                       app f' a'
    Lam _ x t b  -> do t' <- go $ t
                       b' <- goWith x (erase t) b
                       let {-- s1 = tmType t' --}
                           tb = tmType b'
                           lt = norm $ pi x (erase t) (erase tb)
                       -- make sure lt is legal
                       typecheck env lt
                       return $ Lam lt x t' b'

    Quote _ e    -> do te <- go e
                       tt <- typecheck env $ tmType te
                       when (not $ isStar $ tmType tt) $ Left $
                         "typecheck: Quotation is not a term: " ++ show te
                       let failure = Left $ unlines [ 
                                        "typecheck Quote: typedSortInf failed" 
                                      , "t = " ++ show (erase tt)
                                      ]  
                       st <- maybe failure return $ typedSortInf env $ norm tt
                       
                       let t = App () _Exp $ substReplib $ quoteType st
                           t' = norm t
                           failure msg = Left $ unlines [
                                            "typecheck: type of quotation does not kind check:"
                                          , msg  
                                          ]  
                       either failure return $ do
                         typecheck env t
                         typecheck env t'
                       return $ Quote t' te
    Star _       -> return $ Star (Box ())
    Box _        -> return $ Box (Delta ())
                       
-- convert a typed Tm to a sorted Tm
typedSortInf :: Show a => Env -> Tm a -> Maybe Sorted
typedSortInf env t = do
  env' <- envToSortedEnv env
  sortInf env' t

envToSortedEnv :: Env -> Maybe Sorted.Env
envToSortedEnv [] = return []

envToSortedEnv ((x,t):env) = do
  env' <- envToSortedEnv env
  t'   <- sortInf env' t
  return $ (x, classifiedBy $ sort t'):env'

replibEnv = 
  let a = var "a"
      b = var "b"
      _V = var "V"
      _PExp = var "PExp"
      _PExpV = app _PExp _V
      _Exp = var "Exp"
      _ProdStar = var "ProdStar"
      _ProdBox = var "ProdBox"
      _ProdDelta = var "ProdDelta"
      _X = var "X"
      app = S.app
  in  
  [("mkRep", pi "V" (vU ==> star) $ pi "a" vU $ app _V a ==> app _PExpV a)
  ,("mkAbsStar", 
    pi "V" (vU ==> star) $ pi "a" vU $ pi "b" vU $
      (app _V a ==> app2 _PExp _V b) ==> app2 _PExp _V (app2 _ProdStar a b))
  ,("mkAppStar", 
    pi "V" (vU ==> star) $ pi "a" vU $ pi "b" vU $
      app _PExpV (app2 _ProdStar a b) ==> app _PExpV a ==> app _PExpV b)
  ,("mkAbsBox", 
    pi "V" (vU ==> star) $ pi "X" box $ pi "a" (_X ==> vU) $
      (pi "b" _X $ app _PExpV (app a b)) ==> app _PExpV (app2 _ProdBox _X a))
  ,("mkAppBox", 
    pi "V" (vU ==> star) $ pi "X" box $ pi "a" (_X ==> vU) $
      app _PExpV (app2 _ProdBox _X a) ==> (pi "b" _X $ app _PExpV (app a b)))
  ,("mkAbsDelta", 
    pi "V" (vU ==> star) $ pi "a" (box ==> vU) $
      (pi "X" box $ app _PExpV (app a _X)) ==> app _PExpV (app _ProdDelta a))
  ,("mkAppDelta", 
    pi "V" (vU ==> star) $ pi "a" (box ==> vU) $
      app _PExpV (app _ProdDelta a) ==> (pi "X" box $ app _PExpV (app a _X)))
  ,("Var", star ==> vU)
  ,("ProdStar", vU ==> vU ==> vU)
  ,("ProdBox", pi "X" box $ (var "X" ==> vU) ==> vU) 
  ,("ProdDelta", (box ==> vU) ==> vU)
  ,("Exp", vU ==> star)
  ,("PExp", (vU ==> star) ==> vU ==> star)
  ,("U", box) 
  ]
                
  
coerceEnv = [     
    ("coerceExp_*", B._CoerceExpStar)             
  , ("coerceExp_#", B._CoerceExpBox)                          
  , ("coerceExp_^", B._CoerceExpDelta)
  , ("reifyVar",    B._ReifyVar)  
  , ("reifyProd_*", B._ReifyProdStar)
  , ("reifyProd_#", B._ReifyProdBox)  
  , ("reifyProd_^", B._ReifyProdDelta)      
  , ("reflectProd_*", B._ReflectProdStar)
  , ("reflectProd_#", B._ReflectProdBox)  
  , ("reflectProd_^", B._ReflectProdDelta)      
  ]     

coercelib = [     
    ("coerceExp_*", B.coerceExpStar)             
  , ("coerceExp_#", B.coerceExpBox)                          
  , ("coerceExp_^", B.coerceExpDelta)
  , ("reifyVar",    B.reifyVar)  
  , ("reifyProd_*", B.reifyProdStar)
  , ("reifyProd_#", B.reifyProdBox)  
  , ("reifyProd_^", B.reifyProdDelta)      
  , ("reflectProd_*", B.reflectProdStar)
  , ("reflectProd_#", B.reflectProdBox)  
  , ("reflectProd_^", B.reflectProdDelta)      
  ]     


substReplib e = foldr (\(nm,tm) e -> unsafePerformIO $ subst nm tm e) e replib
substCoercelib e = foldr (\(nm,tm) e -> unsafePerformIO $ subst nm tm e) e coercelib