module Language.SystemU.Test.Coerce where

import Language.SystemU.Coerce
import Language.SystemU.Syntax hiding (Var)
import Language.SystemU.Sorted
import Language.SystemU.TypeCheck hiding (app)
import qualified Language.SystemU.Builtins as B

import Test.Framework (testGroup)
import Test.Framework.Providers.HUnit
import Test.HUnit

import Prelude hiding (pi)

tests = testGroup "Coerce" [
    uTests
  , substUTests
  , reifySimpleTests
  , reifyTests  
  , typecheckCoercionTests1
  , typecheckCoercionTests2     
  , coercionFileTests
  ]  

pS = ProdStar
pB = ProdBox
pD = ProdDelta
rS = ReifyProdStar
fS = ReflectProdStar
i  = Identity

[a,b,c] = map var $ words "a b c"
[aU,bU,cU] = map Var [a,b,c]

uTests =
  let sEnv = [("a",Type), ("b",Type), ("c",Type)]
      tEnv = [("a",star), ("b",star), ("c",star)]
      go nm t = testCase nm $ do
                  let Just s = sortInf sEnv t
                      u = tyU s
                      c = compileU u
                      t' = unU u
                      Right k = fmap tmType $ typecheck tEnv c
                  t' @?= t
                  k  @?= B.u
                  
  in testGroup "tyU/unU/compileU" [
    go "a" a
  , go "a -> b" (a ==> b)
  , go "a -> b -> c" (a ==> b ==> c)
  , go "(a -> b) -> c" ((a ==> b) ==> c)
  , go "Pi a:*. a" (pi "a" star a)  
  , go "Pi X:#. a" (pi "X" box a)  
  ]  


substUTests = 
  let go nm u r = testCase nm $ do
                    u' <- substU "a" b u 
                    u' @?= r
    
  in testGroup "substU" [
    go "var" (Var a) (Var b)
  , go "ProdStar 1" (ProdStar (Var a) (Var c))
                    (ProdStar (Var b) (Var c))
  , go "ProdStar 2" (ProdStar (Var c) (Var a))
                    (ProdStar (Var c) (Var b))
  , go "ProdBox kind" (ProdBox "a" a (Var c))  
                      (ProdBox "a" b (Var c))
  , go "ProdBox body" (ProdBox "x" c (Var a))  
                      (ProdBox "x" c (Var b))
  , go "ProdBox shadowed 1" 
                      (ProdBox "a" c (Var a))  
                      (ProdBox "a" c (Var a))
  , go "ProdBox shadowed 2" 
                      (ProdBox "a" a (Var a))  
                      (ProdBox "a" b (Var a))
  ]  

reifySimpleTests = 
  let go nm u r = testCase nm $ reify u @?= r
  in testGroup "reifySimple" [
    -- var
    go "a" aU (i a)
    -- (*,*)
  , go "a -> b" (pS aU bU) (rS aU bU (i a) (i b))
    -- covariance
  , go "a -> b -> c" (pS aU (pS bU cU))  
                     (rS aU (pS bU cU) (i a) (rS bU cU (i b) (i c)))
    -- contravariance
  , go "(a -> b) -> c" (pS (pS aU bU) cU)
                       (rS (pS aU bU) cU
                           (fS aU bU (i a) (i b)) (i c))
  ]

reifyTests =  
  let tmp = ()
  in testGroup "reify" [  
    
  ]  
     
     
-- just reifiers (left hand side is always Var)
typecheckCoercionTests1 =     
  let env = [ ("a",star), ("b",star), ("c",star)
            , ("V", B.u ==> star)
            ] ++ coerceEnv
      v = var "V"
      go nm u1 u2 = testCase nm $ do 
                      let c = coerce u1 u2
                          Right (u1',u2') = typecheckC c
                      u1' @?= u1
                      u2' @?= u2
                      
                      -- now compile coercion and typecheck again
                      let tm = compileCoercion v c
                          t1 = compileU u1
                          t2 = compileU u2
                          r  = norm $ app2 B._PExp v t1 ==> app2 B._PExp v t2
                      tm' <- either fail return $ typecheck env tm
                      d <- runDiff (tmType tm') r
                      d @?= Right ()
                      
                      -- now substitute definitions and typecheck again
                      
  in testGroup "typecheck coercions" [  
    -- variables
    go "reify var" aU aU
  , go "reflect var" (pS aU bU) (pS aU bU)

  , go "reify (*,*)" (Var (a ==> b)) (pS aU bU)
  , go "reflect (*,*)"
       (Var $ (a ==> b) ==> c) (pS (pS aU bU) cU)
  , go "reify (#,*)"
       (Var $ pi "x" star a) (pB "x" star aU)
  , go "reflect (#,*)"
       (Var $ pi "x" star a ==> b) (pS (pB "x" star aU) bU)
  , go "reify (^,*)"
       (Var $ pi "X" box a) (pD "x" aU)
  , go "reflect (^,*)"
       (Var $ pi "X" box a ==> b) (pS (pD "x" aU) bU)    
  ]
     
typecheckCoercionTests2 = 
  let go nm ctx = testCase nm $ do
                    let u1 = ctx (Var (a ==> b))
                        u2 = ctx (ProdStar aU bU)
                        Right (u1',u2') = typecheckC (coerce u1 u2)
                    u1' @?= u1    
                    u2' @?= u2
  in testGroup "typecheck congruence coercions" [
    go "(*,*) covariance" $ \u -> ProdStar aU u
  , go "(*,*) contravariance" $ \u -> ProdStar u aU
  , go "(#,*) covariance" $ \u -> ProdBox "c" star u
  , go "(^,*) covariance" $ \u -> ProdDelta "X" u
  ]
     
coercionFileTests =      
  let mkU1 t a = do
        -- quote, then substitute
        let Just (Pi _ nm (Star _) b) = sortInf [] t
        substU nm a (tyU b)

      mkU2 (Pi _ nm (Star _) b) a = do
        -- substitute, then quote
        b1 <- subst nm a b
        let Just b2 = sortInf [("a",Type), ("b",Type)] b1
        return $ tyU b2
        
      go nm = testCase nm $ do
                tm <- fmap parse $ readFile ("examples/pgms/" ++ nm)
                let Right typed = typecheck [] tm
                    ty = tmType typed
                u1 <- mkU1 ty (a ==> b)
                u2 <- mkU2 ty (a ==> b)
                let Right (u1',u2') = typecheckC (coerce u1 u2)
                u1' @?= u1    
                u2' @?= u2
  in testGroup "coercion file tests" [              
    go "coercionTest1"
  , go "coercionTest2"    
  ]