module Language.SystemU.Test.Syntax where

import qualified Language.SystemU.Core as C
import Language.SystemU.Syntax
import Language.SystemU.Sorted
import qualified Language.SystemU.Builtins as B
import Language.SystemU.QuoteType (expandTypeQuotes, quoteType)
import Language.SystemU.QuoteTerm (expandTermQuotes)
import Language.SystemU.TypeCheck (typecheckTypes, typecheck, tmType, replibEnv, coerceEnv, substReplib, substCoercelib)
import Prelude hiding (pi)

import Test.Framework (testGroup)
import Test.Framework.Providers.HUnit
import Test.HUnit

import Control.Applicative
import System.IO.Unsafe

fromJust (Just x) = x
fromLeft (Left msg) = error msg
fromRight (Right x) = x
isRight (Right _) = True
isRight (Left msg) = error msg

tests = testGroup "Syntax" [
    renameTests
  
  , substTests
  , normTests
  , sortTests  
  , typeExpansionTests
    
  , typecheckTests
  , typecheckBuiltins    
  , expandTypecheckTests  

  , typecheckCpslibTests
  , typecheckCoercelibTests        
  
  , expandTypecheckFileTests    

  , quoteTermTests  
  , quoteTermFileTests

  , betaEquivTests
  ]


renameTests = 
  let go nm i o = testCase nm $ rename "x" "y" i @?= o
   
  in testGroup "rename" [
    go "lam type" (lam "x" (var "x") (var "x")) 
                  (lam "x" (var "y") (var "x"))
  , go "lam body" (lam "a" (var "b") (var "x"))   
                  (lam "a" (var "b") (var "y"))
  , go "lam body shadowed" 
                  (lam "x" (var "a") (var "x"))   
                  (lam "x" (var "a") (var "x"))
  , go "pi type"  (pi "x" (var "x") (var "x")) 
                  (pi "x" (var "y") (var "x"))
  , go "pi body"  (pi "a" (var "b") (var "x")) 
                  (pi "a" (var "b") (var "y"))
  , go "pi body shadowed"  
                  (pi "x" (var "a") (var "x")) 
                  (pi "x" (var "a") (var "x"))
  , go "decl type" (decl "x" (var "x") (var "a") (var "a"))
                   (decl "x" (var "y") (var "a") (var "a"))
  , go "decl def"  (decl "x" (var "a") (var "x") (var "a"))
                   (decl "x" (var "a") (var "y") (var "a"))
  , go "decl body" (decl "a" (var "a") (var "a") (var "x"))
                   (decl "a" (var "a") (var "a") (var "y"))
  , go "decl body shadowed" 
                   (decl "x" (var "a") (var "a") (var "x"))
                   (decl "x" (var "a") (var "a") (var "x"))
  ]

substTests = testGroup "subst" [
    avoidCapture
  ]
             
avoidCapture = testCase "avoid capture" $ do
  s <- show . (+1) <$> nextSymId
  r <- subst "x" (var "y") (lam "y" star $ var "x") 
  r @?= (lam s star $ var "y")
  
-- norm
normTests = 
  let a = var "a"
      f = var "f"
      x = var "x"
      y = var "y"
      go nm tm r = testCase nm $ norm tm @?= r
  in    
  testGroup "norm" [
    -- \\ a:*. \\ x:a. (\\ y:a. y) x
    go "under lambdas"
       (lam "a" star $ lam "x" a $ app (lam "y" a y) x)
       (lam "a" star $ lam "x" a x) 
    -- \\ a:*. \\ x:a. \\ f:a -> a. f ((\\ y:a. y) x)
  , go "on right side of applications"
       (lam "a" star $ lam "x" a $ lam "f" (a ==> a) $ app f (app (lam "y" a y) x))
       (lam "a" star $ lam "x" a $ lam "f" (a ==> a) $ app f x)
  ]

sortTests = testGroup "sortInf" [
    testSortTms
  , testSortTys  
  , testSortKinds  
  ]
            
mkSortTest expected mkGroup =
  let r = Just expected
      check tm = (sort <$> sortInf [] tm) @?= r
      go nm tm = testCase nm (check tm)
      goFile nm = testCase nm $ do
        tm <- fromCore . C.pgm2Tm . read <$> 
          readFile ("examples/pgms/" ++ nm)
        check tm  

  in mkGroup go goFile
  

testSortTms = mkSortTest Term $ \go goFile ->
  let bot = pi "a" star (var "a")
  in    
  testGroup "terms" [
    testGroup "units" [
      go "\\ x:_|_. x" (lam "a" bot (var "a"))  
    , go "\\ a:*. \\ x:a. x" (lam "a" star (lam "x" (var "a") (var "x")))
    , go "\\ X:#. \\ x:_|_. x" (lam "X" box (lam "a" bot (var "a")))
    , go "[\\ x:_|_. x]" (quote (lam "a" bot (var "a")))
    ]
  , testGroup "files" . map goFile . words $
      "mkRep mkAbsStar mkAppStar mkAbsBox mkAppBox mkAbsDelta mkAppDelta "
  ]      
  
testSortTys = mkSortTest Type $ \go goFile ->
  testGroup "types" [
    testGroup "units" [
        go "pi a:*. a" (pi "a" star (var "a"))
      , go "\\ a:*. a" (lam "a" star (var "a"))  
      , go "\\ X:#. pi a:*. a" (lam "X" box (pi "a" star (var "a")))
      , go "(\\ a:*. a) (pi a:*. a)" (app (lam "a" star (var "a")) (pi "a" star (var "a")))
      , go "[pi a:*. a]" (quote (pi "a" star (var "a")))
      ]
  , testGroup "files" . map goFile . words $
      "Var ProdStar ProdBox ProdDelta " ++
      "Rep AbsStar AppStar AbsBox AppBox AbsDelta AppDelta PExp Exp UId"
  ]    
  
testSortKinds = mkSortTest Kind $ \go goFile ->
  testGroup "kinds" [
    go "*" star
  , go "Pi a:*. *" (pi "X" star star)
  , go "Pi X:#. Pi a:X. X" (pi "X" box (pi "a" (var "X") (var "X")))
  , goFile "U"  
  ]

vExp = app B.vExp

-- inline definition
vExpInline t = pi "V" (B.vU ==> star) $ app2 B.vPExp (var "V") t
vVar = app B.vVar
vProdStar = app2 B.vProdStar
vProdBox = app2 B.vProdBox
vProdDelta = app B.vProdDelta

_Exp = app B._Exp
_Var = app B._Var
_ProdStar = app2 B._ProdStar
_ProdBox = app2 B._ProdBox
_ProdDelta = app B._ProdDelta


typeExpansionTests = 
  let go nm ty r = testCase nm $ 
                     (expandTypeQuotes $ fromJust $ sortInf [] ty) @?= norm r
  in    
  testGroup "typeExpansion" [
    go "Pi a:*. [a]"
      (pi "a" star $ quote $ var "a")
      (pi "a" star $ _Var (var "a"))
  , go "pi a:*. [a -> a]"
      (pi "a" star $ quote $ pi "" (var "a") (var "a"))
      (pi "a" star $ _Var (var "a") `_ProdStar` _Var (var "a"))
  , go "[pi a:*. a]"
      (quote $ pi "a" star $ var "a")
      (_ProdBox star $ lam "a" star $ _Var $ var "a")
  , go "pi a:*. [pi X:#.a]"
      (pi "a" star $ quote $ pi "X" box $ var "a")
      (pi "a" star $ _ProdDelta $ lam "X" box $ _Var $ var "a")
  ]

typecheckTests =
  let go nm tm r = testCase nm $ (tmType <$> typecheck [] tm) @?= Right r
  in    
  testGroup "typecheck" [
    go "*" star box
  , go "#" box delta
  , go "\\a:*.a" 
       (lam "a" star $ var "a")
       (pi "a" star star)
  , go "Pi a:*.a" 
       (pi "a" star $ var "a")
       star
  , go "\\a:*. \\x:a. x"     
       (lam "a" star $ lam "x" (var "a") $ var "x")
       (pi "a" star $ pi "" (var "a") $ var "a")
  ]
  
typecheckCpslibTests =
  let env = []
      run tm = isRight $
               fmap tmType $
               typecheck env tm
      go nm = testCase nm $ do
                     tm <- fromCore . C.pgm2Tm . read <$> 
                             readFile ("examples/pgms/" ++ nm)
                     run tm @?= True
  in testGroup "typecheck cpslib" $ 
     map go $ words $ unwords [
       "Ct"  
     , "CPS"  
     , "CPS1"  
     , "DistCPSStar"  
     , "distCPSStar"         
     , "FactorCPSStar"  
     , "factorCPSStar"         
     , "DistCPSBox"  
     , "distCPSBox"         
     , "FactorCPSBox"  
     , "factorCPSBox"         
     , "DistCPSDelta"  
     , "distCPSDelta"         
     , "FactorCPSDelta"         
     , "factorCPSDelta"                
     , "CoerceCPSStar"  
     , "coerceCPSStar"         
     , "CoerceCPSBox"     
     , "coerceCPSBox"            
     , "CoerceCPSDelta"     
     , "coerceCPSDelta"
     ]

typecheckBuiltins =
  let env = []
      run tm = fromRight $
               fmap tmType $
               typecheck env tm
      go nm tm = testCase nm $ do
                   run tm @?= Star ()
  in testGroup "typecheck builtins" $ [
    go "CoerceExpStar" B._CoerceExpStar             
  , go "CoerceExpBox" B._CoerceExpBox                          
  , go "CoerceExpDelta" B._CoerceExpDelta
  , go "ReifyVar" B._ReifyVar  
  , go "ReifyProdStar" B._ReifyProdStar
  , go "ReifyProdBox" B._ReifyProdBox  
  , go "ReifyProdDelta" B._ReifyProdDelta      
  , go "ReflectProdStar" B._ReflectProdStar
  , go "ReflectProdBox" B._ReflectProdBox  
  , go "ReflectProdDelta" B._ReflectProdDelta      
  ]     



typecheckCoercelibTests =
  let env = [("distExp_*", B._DistExpStar)
            ,("distExp_#", B._DistExpBox) 
            ,("distExp_^", B._DistExpDelta) 
            ,("factorExp_*", B._FactorExpStar)
            ,("factorExp_#", B._FactorExpBox)             
            ,("factorExp_^", B._FactorExpDelta)
            ,("coerceExp_*", B._CoerceExpStar)             
            ,("coerceExp_#", B._CoerceExpBox)                          
            ,("coerceExp_^", B._CoerceExpDelta)
            ,("_BOTTOM_", pi "x" star $ var "x") 
            ]  ++ replibEnv
      run tm = isRight $
               fmap tmType $
               typecheck env tm
      go nm = testCase nm $ do
                     tm <- fromCore . C.pgm2Tm . read <$> 
                             readFile ("examples/pgms/" ++ nm)
                     run tm @?= True
  in testGroup "typecheck coercelib" $ 
     map go $ words $ unwords [
       -- kind check Dist_*/#/^
       "testDistStar"  

     , "testDistBox"  
     , "testDistDelta"  
       
       -- kind check DistExp_*/#/^
     , "testDistExpStar"  
     , "testDistExpBox"  
     , "testDistExpDelta"  
       
       
     , "distRStar"  
       
       -- type check dist_*/#/^
     , "distExpStar" 
     , "distExpBox"  
     , "distExpDelta"  
       
       -- typecheck factor_*/#/^
     , "factorExpStar"  
     , "factorExpBox"  
     , "factorExpDelta"  
       
       -- typecheck coerceProd_*/#/^
     , "coerceExpStar"  
     , "coerceExpBox"  
     , "coerceExpDelta"
       
     , "reifyVar"
     , "reflectVar"  
       
     , "reifyProdStar"
     , "reflectProdStar"  

     , "ReifyProdBox"
     , "reifyProdBox"
     , "ReflectProdBox"
     , "reflectProdBox"  
     
     , "ReifyProdDelta"
     , "reifyProdDelta"
     , "ReflectProdDelta"
     , "reflectProdDelta"  
       
     , "reify1"  
     , "reflect1"  
     , "coerce1"  

     , "testCoerceExpStar"         
     , "testCoerceExpBox"  
     , "testCoerceExpBox2"
     , "testCoerceExpBox3"       
     ]
     
expandTypecheckTests =
  let run tm = fmap tmType $
               typecheck replibEnv $ 
               expandTypeQuotes $ 
               fromJust $ sortInf [] tm
      go nm tm r = testCase nm $ run tm @?= Right (norm r)
  in      
  testGroup "typecheck . expandTypeQuotes" [
    go "*" star box
  , go "#" box delta
  , go "\\a:*.a" 
       (lam "a" star $ var "a")
       (pi "a" star star)
  , go "Pi a:*.a" 
       (pi "a" star $ var "a")
       star
  , go "\\a:*. \\x:a. x"     
       (lam "a" star $ lam "x" (var "a") $ var "x")
       (pi "a" star $ pi "" (var "a") $ var "a")
  , go "(\\X:#. \\a:X. a) (* -> *) (\\b:*. b)"
       (let f = lam "X" box $ lam "a" (var "X") $ var "a"
            k = pi "" star star
            t = lam "b" star (var "b")
        in app2 f k t)
       (pi "" star star)
  , go "[Pi a:*.a]"     
       (quote $ pi "a" star $ var "a")
       B.u
  , go "[Pi a:*.a -> a]"     
       (quote $ pi "a" star $ pi "" (var "a") $ var "a")
       B.u
  , go "[Pi X:#. Pi a:*. a -> a]"     
       (quote $ pi "X" box $ pi "a" star $ pi "" (var "a") $ var "a")
       B.u
  , go "[\\a:*. \\x:a. x]"   
       (quote $ lam "a" star $ lam "x" (var "a") $ var "x")
       (_Exp $ _ProdBox star $ lam "a" star $
                      _ProdStar (_Var $ var "a") (_Var $ var "a"))
  , go "[\\ X:#. \\a:*. \\x:a. x]"   
       (quote $ lam "X" box $ lam "a" star $ lam "x" (var "a") $ var "x")
       (_Exp $ _ProdDelta $ lam "X" box $
                  _ProdBox star $ lam "a" star $
                      _ProdStar (_Var $ var "a") (_Var $ var "a"))
  , go "[\\a:*. \\x:a -> a. \\y:a. x y]"   
       (quote $ lam "a" star $ 
                lam "x" (pi "" (var "a") (var "a")) $ 
                lam "y" (var "a") $
                app (var "x") (var "y"))
       (let a = _Var $ var "a"
            a2a = _ProdStar a a
        in   
        _Exp $ _ProdBox star $ lam "a" star $
               _ProdStar a2a a2a)
  , go "[\\a:*. \\x:a -> a. \\y:a. x y]"   
       (quote $ lam "a" star $ 
                lam "x" (pi "" (var "a") (var "a")) $ 
                lam "y" (var "a") $
                app (var "x") (var "y"))
       (let a = _Var $ var "a"
            a2a = _ProdStar a a
        in   
        _Exp $ _ProdBox star $ lam "a" star $
               _ProdStar a2a a2a)
  ]
  
expandTypecheckFileTests =
  let env = [("_BOTTOM_", pi "x" star $ var "x")] ++ replibEnv
      run tm = isRight $
               fmap tmType $
               typecheck env $ 
               expandTypeQuotes $ 
               fromJust $ sortInf [("_BOTTOM_", Type)] tm
      go nm = testCase nm $ do
                     tm <- fromCore . C.pgm2Tm . read <$> 
                             readFile ("examples/pgms/" ++ nm)
                     run tm @?= True
  in testGroup "Files: (typecheck . expandTypeQuotes)" $ 
     map go $ words $ unwords [
        "ProdStar ProdBox ProdDelta "
      , "Rep AbsStar AppStar AbsBox AppBox AbsDelta AppDelta " 
      , "PExp Exp "
      , "mkRep mkAbsStar mkAppStar mkAbsBox mkAppBox mkAbsDelta mkAppDelta "
      , "tcMkRep tcMkAbsStar tcMkAppStar tcMkAbsBox tcMkAppBox tcMkAbsDelta tcMkAppDelta"
      , "UId tcUId"
      , "idRep idStar idStar2 idBox idBox2 idDelta idDelta2 "
      , "unquote"  
      , "typeCPS"
      , "typeIsAbs"
      , "testUnquote"
      ]  

loadProgram nm = parse <$> readFile ("examples/pgms/" ++ nm)
loadExample nm = parse <$> readFile ("examples/" ++ nm)

runEquivTest :: Tm () -> IO (Tm (), Tm ())
runEquivTest = go
  where
      [a,x,y] = map var $ words "a x y"
      fst = lam "a" star $ lam "x" a $ lam "y" a x
      snd = lam "a" star $ lam "x" a $ lam "y" a y
      die phase msg = error $ unlines [phase,msg]
      run tm = either (die "typecheck 1") id $               
               typecheck [] $
               substCoercelib $
               substReplib $
               either (die "expandTermQuotes") id $
               expandTermQuotes $
               either (die "typecheck 1") id $               
               typecheck replibEnv $ 
               expandTypeQuotes $ 
               maybe (error "sortInf failed") id $ 
               sortInf [] tm
      go tm = do let tm' = erase $ run tm        
                     a = app tm' fst
                     b = app tm' snd
                 return (eta $ norm a, eta $ norm b)

betaEquivTests =
  let go nm = testCase nm $ do
                tm <- loadProgram nm
                (a,b) <- runEquivTest tm
                a @?= b

  in testGroup "beta equivalence tests" [
    -- type decomposition tests: Thm 3.4
    go "decomposePiStar"
  , go "decomposePiBox"
  , go "decomposePiDelta"  
    
    -- correctness of UId: Thm 4.5
  , go "testUId"
    
    -- type representation decomposition tests: Thm 4.7
  , go "decomposeProdStar"
  , go "decomposeProdBox"
  , go "decomposeProdDelta"  
    
{--
    -- unquote functional tests
    -- subsumed by the self-application test below
    go "unquoteAbsStar"
  , go "unquoteAppStar"    
  , go "unquoteAbsBox"    
  , go "unquoteAppBox"    
  , go "unquoteAppBox2"    
  , go "unquoteAppBox3"    
  , go "unquoteAppBox4"      
  , go "unquoteAbsDelta"    
  , go "unquoteAppDelta"    

  , go "unquoteIdRep"
  , go "unquoteIdStar"
  , go "unquoteIdStar2"
  , go "unquoteIdBox"
  , go "unquoteIdBox2"
  , go "unquoteIdBox"
  , go "unquoteIdBox2"
  , go "unquoteIdDelta"
  , go "unquoteIdDelta2"

  , go "unquoteWitnessUId"

  , go "unquoteUnquoteEquiv2"
--}
    -- self-application 
  , go "selfApplyUnquote"     
  , go "selfApplyIsAbs"

    -- CPS functional tests
  , go "cpsPolyappStar"
  , go "cpsPolyappStar2"
   
  , go "cpsPolyappBox"    
  , go "cpsPolyappDelta"        
  ]
     

quoteEnv = replibEnv ++ coerceEnv

typecheckQuoteVar tm = typecheck replibEnv tm >>=
                       expandTermQuotes >>=
                       typecheck quoteEnv -- . unsafePerformIO . subst "U" B.u 


typecheckQuoteDef tm = do
  tm1 <- maybe (Left "sortInf failed") Right $ sortInf [] tm
  let tm2 = expandTypeQuotes tm1
  (typecheck replibEnv tm2 >>=
   expandTermQuotes >>=
   typecheck coerceEnv . substReplib)

quoteTermTests =
  let a = var "a"
      b = var "b"
      f = var "f"
      x = var "x"
      vExp = vExpInline
      {--
      go nm tm r = testGroup nm [goVar nm tm r, goDef nm tm r]
      goVar nm tm r = testCase "var" $ expect typecheckQuoteVar tm r
      goDef nm tm r = testCase "def" $ expect typecheckQuoteDef tm (substReplib r)
      --}
      go nm tm r = testCase nm $ expect typecheckQuoteDef tm (substReplib r)      
      expect tcQuote tm r = 
        case tcQuote tm of
          Left msg -> putStrLn msg >> fail ""
          Right tm' -> tmType tm' @?= norm r
  in  
  testGroup "quoteTm" [
    go "term abstraction and variable: \\a:*. [\\x:a. x]" 
       (lam "a" star $ quote $ lam "x" a x)
       (pi "a" star $ vExp $ vProdStar (vVar a) (vVar a))
  , go "arrow type: \\a:*. [\\x:a -> a. x]" 
       (lam "a" star $ quote $ lam "x" (a ==> a) x)
       (pi "a" star $ vExp $ vProdStar (vProdStar (vVar a) (vVar a))
                                       (vProdStar (vVar a) (vVar a)))
  , go "term application: \\a:*. [\\f:a -> a. \\x:a. f x]" 
       (lam "a" star $ quote $ lam "f" (a ==> a) $ lam "x" a $ app f x)
       (pi "a" star $ vExp $ 
          vProdStar (vProdStar (vVar a) (vVar a))
                    (vProdStar (vVar a) (vVar a)))
  , go "type abstraction: [\\a:*. \\x:a. x]"
       (quote $ lam "a" star $ lam "x" a x)
       (vExp $ vProdBox star $ lam "a" star $ vProdStar (vVar a) (vVar a))
  , go "type application: \\a:*. [\\f: (Pi b:*.b). f a]"
       (lam "a" star $ quote $ lam "f" (pi "b" star b) $ app f a)
       (pi "a" star $ vExp $ 
          vProdStar (vProdBox star $ lam "b" star $ vVar b) (vVar a))
  , go "kind abstraction: \\a:*. [\\X:#. \\x:a. x]"
       (lam "a" star $ quote $ lam "X" box $ lam "x" a x)
       (pi "a" star $ vExp $ 
          vProdDelta $ lam "X" box $ vProdStar (vVar a) (vVar a))
  , go "kind application: \\a:*. [\\f:(Pi b:#. a). f *]"
       (lam "a" star $ quote $ lam "f" (pi "b" box a) $ app f star)
       (pi "a" star $ vExp $ 
          vProdStar (vProdDelta $ lam "b" box $ vVar a) (vVar a))
  ]  

quoteTermFileTests = 
  let {--
      go nm    = testGroup nm [goVar nm, goDef nm]
      goVar nm = testCase "var" $ run typecheckQuoteVar nm
      goDef nm = testCase "def" $ run typecheckQuoteDef nm
      --}
      go nm = testCase nm $ run typecheckQuoteDef nm    
      run tcQuote nm = do tm <- loadExample nm
                          isRight (tcQuote tm) @?= True
  in        
  testGroup "quoteTm files" [
    go "twice"
  , go "polyappStar"  
  , go "polyappBox"  
  , go "polyappDelta"  
  , go "pgms/kindsOfTypeConstructors"  
  , go "pgms/kindsOfTypeRepresentations"  
  , go "pgms/testCoercion"  
  , go "pgms/testCoercionEtaExpanded"
  , go "pgms/typeUnquote"        
  , go "pgms/typeIsAbs"    
  , go "pgms/testUnquote"  
  , go "pgms/unquoteUnquote"  
  , go "pgms/typeCPS"
  , go "pgms/testCPS"  
  , go "pgms/selfApplyCPS"  
  ]
